from django.db import models

from tom_targets.models import Target


class Superevent(models.Model):
    """Represents a Superevent being followed-up upon by this TOM.

    A Superevent is distinguished from a Target in that it is localized to a region of the sky
    (vs. a specific RA,DEC). The potential Targets in the localization region must be identified,
    prioritized, and categorized (retired, of-interest, etc) for follow-up EM observations

    For the moment, this is rather GraceDB (GW) specific, but sh/could be generalized to work
    with gamma-ray burst and neutrino events.
    """

    class SupereventType(models.TextChoices):
        GRAVITATIONAL_WAVE = 'GW', 'Gravitational Wave'
        GAMMA_RAY_BURST = 'GRB', 'Gamma-ray Burst'
        NEUTRINO = 'NU', 'Neutrino'
        UNKNOWN = 'UNK', 'Unknown'

    superevent_type = models.CharField(
        max_length=3,
        choices=SupereventType.choices,
        default=SupereventType.GRAVITATIONAL_WAVE,
    )

    # TODO: ask Curtis/Rachel/Andy about generalized use cases.
    superevent_id = models.CharField(max_length=64)  # GraceDB superevent_id reference
    superevent_url = models.URLField()  # TODO: this should instead be constructed via superevent_id

    created = models.DateTimeField(auto_now_add=True)
    modified = models.DateTimeField(auto_now=True)

    @property
    def gracedb_url(self):
        """Construct and return the GraceDB URL for this superevent from the superevent_id Field.

        for example, https://gracedb.ligo.org/superevents/S200316bj/
        """
        # TODO: add check that superevent_type is GRAVITATIONAL_WAVE
        return f"https://gracedb.ligo.org/superevents/{self.superevent_id}/"

    @property
    def treasuremap_url(self):
        """Construct and return the Treasure Map (treasuremap.space) URL for this superevent
        from the superevent_id Field.

        for example: http://treasuremap.space/alerts?graceids=S200219ac
        """
        # TODO: add check that superevent_type is GRAVITATIONAL_WAVE
        return f"http://treasuremap.space/alerts?graceids={self.superevent_id}"

    def __str__(self):
        return self.superevent_id


class EventCandidate(models.Model):
    target = models.ForeignKey(Target, on_delete=models.CASCADE)
    superevent = models.ForeignKey(Superevent, on_delete=models.CASCADE)

    viable = models.BooleanField(
        default=True,
        # TODO: add description, etc
    )
    priority = models.IntegerField(
        default=1,
        # TODO: add description, etc
    )

    class Meta:
        constraints = [  # TODO: this constraint isn't working
            models.UniqueConstraint(fields=['target', 'superevent'], name='Unique Target/Superevent')
        ]

    def __str__(self):
        return f'EventCandidate({self.id}) Superevent: {self.superevent} Target: {self.target}'


class EventLocalization(models.Model):
    """Represents a region of the sky in which a superevent may have taken place.
    """
    created = models.DateTimeField(auto_now_add=True)
    modified = models.DateTimeField(auto_now=True)



from django.contrib import admin

from .models import EventCandidate, EventLocalization, Superevent

# Register your models here.
admin.site.register(EventCandidate)
admin.site.register(EventLocalization)
admin.site.register(Superevent)



import factory

from tom_nonlocalizedevents.models import Superevent, EventLocalization


class SupereventFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = Superevent

    superevent_id = factory.Faker('pystr')
    superevent_url = factory.Faker('pystr')


class EventLocalizationFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = EventLocalization



from django.contrib.auth.models import User
from django.urls import reverse
from rest_framework.test import APITestCase

from tom_nonlocalizedevents.tests.factories import SupereventFactory, EventLocalizationFactory


class SupereventAPITestCase(APITestCase):
    def setUp(self):
        self.user = User.objects.create(username='test_user')
        self.superevent1 = SupereventFactory.create(superevent_id='superevent1')
        self.superevent2 = SupereventFactory.create(superevent_id='superevent2')
        self.eventlocalization1 = EventLocalizationFactory.create()
        self.eventlocalization2 = EventLocalizationFactory.create()

        # TODO: sort out django-guardian permissions
        # assign_perm('tom_targets.view_target', self.user, self.st2)

        self.client.force_login(self.user)


class TestSupereventViewSet(SupereventAPITestCase):

    def test_superevent_list(self):
        """Test Superevent API list endpoint."""
        response = self.client.get(reverse('api:superevent-list'))

        self.assertEqual(response.json()['count'], 2)
        self.assertContains(response, f'"superevent_id":"{self.superevent1.superevent_id}"')
        self.assertContains(response, f'"superevent_url":"{self.superevent1.superevent_url}"')
        self.assertContains(response, f'"superevent_id":"{self.superevent2.superevent_id}"')
        self.assertContains(response, f'"superevent_url":"{self.superevent2.superevent_url}"')


class TestEventLocalizationViewSet(SupereventAPITestCase):
    def test_eventlocalization_list(self):
        """Test EventLocalization API list endpoint."""
        response = self.client.get(reverse('api:eventlocalization-list'))

        self.assertEqual(response.json()['count'], 2)



import json

from django.contrib import messages
from django.core.cache import cache
from django.shortcuts import redirect
from django.views.generic import DetailView, ListView
from django.views.generic.base import View
from django.urls import reverse

from rest_framework import permissions, viewsets

# from tom_alerts.alerts import get_service_class

from tom_nonlocalizedevents.superevent_clients.gravitational_wave import GravitationalWaveClient

from .models import EventCandidate, EventLocalization, Superevent
from .serializers import EventCandidateSerializer, EventLocalizationSerializer, SupereventSerializer


class NonlocalizedEventListView(ListView):
    """
    Unadorned Django ListView subclass for Superevent model.
    (To be updated when Superevent model is renamed to NonlocalizedEvent).
    """
    model = Superevent
    template_name = 'tom_nonlocalizedevents/index.html'


class NonlocalizedEventDetailView(DetailView):
    """
    Django DetailView subclass for SuperEvent model.
    (To be updated when Superevent model is renamed to NonlocalizedEvent).

    Has mechanism to supply templates specific to the type of NonlocalizedEvent
    (GW, GRB, Nutrino).
    """
    model = Superevent
    template_name = 'tom_nonlocalizedevents/detail.html'

    # TODO: consider combining these dictionaries
    template_mapping = {
        Superevent.SupereventType.GRAVITATIONAL_WAVE:
            'tom_nonlocalizedevents/superevent_detail/gravitational_wave.html',
        Superevent.SupereventType.GAMMA_RAY_BURST:
            'tom_nonlocalizedevents/superevent_detail/gamma_ray_burst.html',
        Superevent.SupereventType.NEUTRINO:
            'tom_nonlocalizedevents/superevent_detail/neutrino.html',
    }

    # A client in this context is the interface to the service providing event info.
    # (i.e GraceDB for gravitational wave events)
    client_mapping = {
        Superevent.SupereventType.GRAVITATIONAL_WAVE: GravitationalWaveClient(),
        Superevent.SupereventType.GAMMA_RAY_BURST: None,
        Superevent.SupereventType.NEUTRINO: None,
        Superevent.SupereventType.UNKNOWN: None,
    }

    def get_template_names(self):
        obj = self.get_object()
        return [self.template_mapping[obj.superevent_type]]

    # TODO: error handling
    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        obj = self.get_object()
        superevent_client = self.client_mapping[obj.superevent_type]
        # TODO: should define superevent_client API (via ABC) for clients to implement
        if superevent_client is not None:
            context['superevent_data'] = superevent_client.get_superevent_data(obj.superevent_id)
            context.update(superevent_client.get_additional_context_data(obj.superevent_id))
        return context


# from the tom_alerts query_result.html

class CreateEventFromSCiMMAAlertView(View):
    """
    Creates the models.Superevent instance and redirect to NonlocalizedEventDetailView
    """
    pass

    def post(self, request, *args, **kwargs):
        """
        """
        # the request.POST is a QueryDict object;
        # for SCiMMA, alerts: list items are PKs to skip.dev.hop.scimma.org/api/alerts/PK/
        query_id = self.request.POST['query_id']
        # broker_name = self.request.POST['broker']  # should be "SCIMMA"
        # broker_class = get_service_class(broker_name)  # should be <class 'tom_scimma.scimma.SCIMMABroker'>
        alerts = [int(id) for id in request.POST.getlist('alerts', [])]

        errors = []
        # if the user didn't select an alert; warn and re-direct back
        if not alerts:
            messages.warning(request, 'Please select at least one alert from which to create an event.')
            reverse_url: str = reverse('tom_alerts:run', kwargs={'pk': query_id})
            return redirect(reverse_url)

        # try to extract EventID from Alert and use it to create a Superevent
        for alert_id in alerts:
            cached_alert = json.loads(cache.get(f'alert_{alert_id}'))  # cached by tom_alerts.views.py::RunQueryView
            # early return: alert not in cache
            if not cached_alert:
                messages.error(request, 'Could not create event(s). Try re-running the query to refresh the cache.')
                return redirect(reverse('tom_alerts:run', kwargs={'pk': query_id}))

            # early return: alert not LVC/LVC COUNTERPART NOTICE
            if cached_alert.get('topic', '') != 'lvc.lvc-counterpart':
                messages.error(request,
                               ('Only Alerts from the lvc.lvc-counterpart topic have '
                                'parsed event_trig_num required for Event origination. '
                                'Please select an appropriate alert.'))
                return redirect(reverse('tom_alerts:run', kwargs={'pk': query_id}))

            # early return: event_trig_num not found in parsed alert message
            event_trig_num = cached_alert['message'].get('event_trig_num', None)
            if event_trig_num is None:
                messages.error(request,
                               (f'Could not create event for alert: {alert_id}. '
                                'event_trig_num not found in alert message.'))
                return redirect(reverse('tom_alerts:run', kwargs={'pk': query_id}))

            superevent, created = Superevent.objects.get_or_create(superevent_id=event_trig_num)
            if not created:
                # the superevent already existed
                messages.warning(request, f'Event {event_trig_num} already exists.')
                errors.append(superevent.superevent_id)

        if len(alerts) == len(errors):
            # zero superevents created
            return redirect(reverse('tom_alerts:run', kwargs={'pk': query_id}))
        elif len(alerts) == 1:
            # one superevent created
            return redirect(reverse('nonlocalizedevents:detail', kwargs={'pk': superevent.id}))
        else:
            # multipe superevents created
            return redirect(reverse('nonlocalizedevents:index'))

# Django Rest Framework Views


class NonlocalizedEventViewSet(viewsets.ModelViewSet):
    """
    DRF API endpoint that allows Superevents to be viewed or edited.
    """
    queryset = Superevent.objects.all()
    serializer_class = SupereventSerializer
    permission_classes = []


class EventCandidateViewSet(viewsets.ModelViewSet):
    """
    DRF API endpoint for EventCandidate model.

    Implementation has changes for bulk_create and update/PATCH EventCandidate instances.
    """
    queryset = EventCandidate.objects.all()
    serializer_class = EventCandidateSerializer
    permission_classes = []  # TODO: re-implement auth permissions

    def get_serializer(self, *args, **kwargs):
        # In order to ensure the list_serializer_class is used for bulk_create, we check that the POST data is a list
        # and add `many = True` to the kwargs
        if isinstance(kwargs.get('data', {}), list):
            kwargs['many'] = True

        return super().get_serializer(*args, **kwargs)

    def update(self, request, *args, **kwargs):
        """Provide support for the PATCH HTTP verb to update individual model fields.

        An example request might look like:

            PATCH http://localhost:8000/api/eventcandidates/18/

        with a Request Body of:

            {
                "viability": false
            }

        """
        kwargs['partial'] = True
        return super().update(request, *args, **kwargs)


class EventLocalizationViewSet(viewsets.ModelViewSet):
    """
    DRF API endpoint that allows EventLocalizations to be viewed or edited.
    """
    queryset = EventLocalization.objects.all()
    serializer_class = EventLocalizationSerializer
    permission_classes = [permissions.IsAuthenticated]



from rest_framework import serializers
from tom_targets.models import Target
from tom_targets.serializers import TargetSerializer

from tom_nonlocalizedevents.models import EventCandidate, EventLocalization, Superevent


class BulkCreateEventCandidateListSerializer(serializers.ListSerializer):
    def create(self, validated_data):
        event_candidates = [EventCandidate(**item) for item in validated_data]
        return EventCandidate.objects.bulk_create(event_candidates)


class EventCandidateSerializer(serializers.ModelSerializer):
    """
    Serializer class for the ``EventCandidate``. ``PrimaryKeyRelatedField``s are used in order to allow creating an
    ``EventCandidate`` with just a primary key, and ``to_representation`` is then overridden for proper display values.
    See: https://www.django-rest-framework.org/api-guide/relations/#custom-relational-fields
    """
    superevent = serializers.PrimaryKeyRelatedField(queryset=Superevent.objects.all())
    target = serializers.PrimaryKeyRelatedField(queryset=Target.objects.all())

    viable = serializers.BooleanField(default=True)
    priority = serializers.IntegerField(default=1)

    class Meta:
        model = EventCandidate
        fields = '__all__'
        list_serializer_class = BulkCreateEventCandidateListSerializer

    def to_representation(self, instance):
        representation = super().to_representation(instance)
        # TODO: this is a little unorthodox. it's convenient for now to handle the data from
        # the ForiengKey objects, but that should be handled separately by their own serializers.
        # (and to_representation could be left undisturbed).
        representation['target'] = TargetSerializer(Target.objects.get(pk=representation['target'])).data
        representation['superevent'] = Superevent.objects.get(pk=representation['superevent']).superevent_id
        return representation


class SupereventSerializer(serializers.HyperlinkedModelSerializer):
    event_candidates = serializers.SerializerMethodField()

    class Meta:
        model = Superevent
        fields = ['superevent_id', 'superevent_url',
                  'id', 'event_candidates', 'created', 'modified']

    def get_event_candidates(self, instance):
        alerts = instance.eventcandidate_set.all()
        # This returns the superevent identifier, which means it's duplicated in the response. The SupereventSerializer
        # should therefore use its own custom EventCandidateSerializer rather than the one defined above
        return EventCandidateSerializer(alerts, many=True).data


class EventLocalizationSerializer(serializers.HyperlinkedModelSerializer):
    class Meta:
        model = EventLocalization
        fields = ['id', 'created', 'modified']
